# -*- coding: utf-8 -*-
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

# For Backward compatibility
from ansible.module_utils.common.sentinel import Sentinel  # pylint: disable=unused-import
