# WARNING: Changing line numbers of code in this file will break collection tests that use tracing to check paths and line numbers.
#          Also, do not import annotations from __future__ as this will break detection of __future__ inheritance.


def question() -> float:
    return 3 / 2
