#SPDX-License-Identifier: MIT-0
#
# (c) 2018 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations
from ansible.errors import AnsibleError

try:
    from ansible.plugins.cliconf import CliconfBase
    """
        Examples:
            https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/cliconf/iosxr.py
            https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/cliconf/junos.py
    """
except ImportError:
    raise AnsibleError("Cliconf Plugin [ {{ role_name }} ]: Dependency not satisfied")


class Cliconf(CliconfBase):
    """
        Examples:
            https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/cliconf/iosxr.py
            https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/cliconf/junos.py
    """
    raise AnsibleError("Cliconf Plugin [ {{ role_name }} ]:  Not implemented")
