using System;

namespace ansible_collections.ns.col.plugins.module_utils.CSharpSigned
{
    public class TestClass
    {
        public static string TestMethod(string input)
        {
            return input;
        }
    }
}
