# -*- coding: utf-8 -*-

# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


class ModuleDocFragment:

    DOCUMENTATION = r"""
options:
  option1:
    description:
    - Test description
    required: yes
    aliases:
    - alias1
    type: str
"""
