# Copyright (c) 2024 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: with_py
short_description: Short description
description: Long description
options: {}
author:
  - Ansible Core Team (@ansible)
"""

EXAMPLES = r"""
- name: Some example
  ns.col.with_py:
"""

RETURNS = ""
